import asyncio
import collections
import logging
import os
import random
import re
import time

from pprint import pformat

from model_ensembler.tasks.utils import execute_command
from model_ensembler.cluster import Job, job_lock
from model_ensembler.utils import Arguments

START_STATES = ("COMPLETING", "PENDING", "RESV_DEL_HOLD", "RUNNING",
                "SUSPENDED", "CONFIGURING", "REQUEUE_FED", "REQUEUE_HOLD",
                "REQUEUED", "RESIZING", "REVOKED", "SIGNALED", "STOPPED",
                "FAILED")
FINISH_STATES = ("COMPLETED", "FAILED", "CANCELLED", "OUT_OF_MEMORY",
                 "DEADLINE", "NODE_FAIL", "PREEMPTED", "TIMEOUT")


async def find_id(job_id):
    """Method to find SLURM job by ID

    This method provides an interface to the squeue SLURM queue utility to
    identify a job and return it along with it's state

    Args:
        job_id (int): SLURM job identifier

    Returns:
        jobs (list): job objects including name and state
    """
    job = None
    args = Arguments()

    while not job:
        try:
            res = await execute_command("sacct -XnP -j {} "
                                        "-o jobname,state,start,end".
                                        format(job_id))
            output = res.stdout.decode().strip()
            (name, state, started, finished) = output.split("|")
        except ValueError:
            logging.warning("Could not retrieve job from list")
            await asyncio.sleep(args.check_timeout)
        else:
            job = Job(
                name=name,
                state=state.split()[0],
                started=started == "Unknown",
                finished=finished == "Unknown"
            )

            logging.debug("SLURM find result name: {}".format(job.name))
    return job


async def current_jobs(ctx, match):
    filtered_jobs = None

    # Ensure we account for empty lists
    while not filtered_jobs and filtered_jobs is None:
        try:
            res = await execute_command("squeue -o \"%j,%T\" -h -p {}".
                                        format(ctx.cluster),
                                        cwd=ctx.dir)
            output = res.stdout.decode()
        except Exception as e:
            logging.warning("Could not retrieve list: {}".format(e))
        else:
            jobs = []
            for line in output.split():
                fields = line.strip().split(",")
                jobs.append({"name": fields[0], "job_state": fields[1]})

            filtered_jobs = [{"name": j['name'], "state": j["job_state"]}
                         for j in jobs
                         if j['name'].startswith(match)
                         and j['job_state'] in START_STATES]

            logging.debug("SLURM JOBS result: {}".
                          format(pformat(filtered_jobs)))

    return filtered_jobs


async def submit_job(ctx, script=None):
    r_sbatch_id = re.compile(r'Submitted batch job (\d+)$')
    args = Arguments()

    # Don't smash the scheduler immediately, it appears to have the potential
    # to cause problems. 
    sleep_for = random.randint(0, args.max_stagger)
    logging.debug("Sleeping for {} seconds before submission".format(sleep_for))
    await asyncio.sleep(sleep_for)
    res = await execute_command("sbatch --no-requeue {}".format(script),
                                cwd=ctx.dir)
    output = res.stdout.decode()

    sbatch_match = r_sbatch_id.match(output)
    if sbatch_match:
        job_id = sbatch_match.group(1)
        logging.info("Submitted job with ID {}".format(job_id))

        return int(job_id)
    return None



"""Task Exceptions

Exceptions relating to tasks
"""


class EnsembleException(Exception):
    """Common ensemble exception for common implementation"""
    pass


class CheckException(EnsembleException):
    """For check failures"""
    pass


class FailureNotToleratedError(RuntimeError):
    """Check failure error cannot be tolerated"""
    pass


class ProcessingException(EnsembleException):
    """For general processing failures"""
    pass


class TaskException(EnsembleException):
    """For task processing failures"""
    pass




from model_ensembler.tasks.exceptions import \
    CheckException, ProcessingException, TaskException
from model_ensembler.tasks.hpc import jobs, quota, submit
from model_ensembler.tasks.sys import check, execute, move, remove

__all__ = [
    "jobs", "quota", "submit",
    "check", "execute", "move", "remove",
    "CheckException", "ProcessingException", "TaskException"
]



import argparse
import logging
import os
import re

from .config import EnsembleConfig
from .batcher import BatchExecutor
# TODO: logging and parse_args should be in utils
from .utils import Arguments, background_fork, setup_logging

"""CLI entry module

This module contains all package entry_points
"""


def parse_indexes(argv):
    """ Method for ensuring a CSV string of integers

    Returns:
        list: List of integer values
    Raises:
        argparse.ArgumentTypeError
    """
    if re.match(r'^([0-9]+,)*[0-9]+$', argv):
        return [int(v) for v in argv.split(",")]
    raise argparse.ArgumentTypeError("{} is not a CSV delimited integer list "
                                     "of indexes".format(argv))


def parse_extra_vars(arg):
    """ Method for processing extra var arguments

    Returns:
        tuple: name and value for the argument to be overridden
    Raises:
        argparse.ArgumentTypeError
    """

    arg_match = re.match(r'^([^=]+)=(.+)$', arg)
    if arg_match:
        return arg_match.groups()
    raise argparse.ArgumentTypeError("Argument does not match "
                                      "name=value format: {}".format(arg))


def parse_args():
    """Parse command line parameters

    Returns:
        Arguments: The immutable instance from ``.utils``
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--daemon",
                        help="Daemonise the ensembler", default=False,
                        action="store_true")

    # TODO: Need to validate the argument selections/group certain commands
    parser.add_argument("-v", "--verbose",
                        help="Log verbosely", default=False,
                        action="store_true")
    parser.add_argument("-c", "--no-checks",
                        help="Do not run check commands", default=False,
                        action="store_true")
    parser.add_argument("-s", "--no-submission",
                        help="Do not try to submit job, just log the step",
                        default=False, action="store_true")
    parser.add_argument("-p", "--pickup",
                        help="Continue a previous set of runs by picking up "
                        "existing directories rather, than assuming to create "
                        "them; for example if ensemble has previously failed",
                        default=False, action="store_true")
    parser.add_argument("-l", "--shell",
                        help="Allows the user to specify the shell passed to "
                             "subprocess execs.",
                        default="/bin/bash", type=str)

    # FIXME: These should not be applied in multi-batch ensembles
    parser.add_argument("-k", "--skips",
                        help="Number of run entries to skip", default=0,
                        type=int)

    parser.add_argument("-i", "--indexes",
                        help="Specify which indexes to run",
                        type=parse_indexes)

    parser.add_argument("-ct", "--check-timeout", default=30, type=int)
    parser.add_argument("-st", "--submit-timeout", default=20, type=int)
    parser.add_argument("-rt", "--running-timeout", default=60, type=int)
    parser.add_argument("-et", "--error-timeout", default=120, type=int)

    parser.add_argument("-ms", "--max-stagger", default=1, type=int)

    parser.add_argument("-x", "--extra-vars", dest="extra", nargs="*",
                        default=[], type=parse_extra_vars)

    parser.add_argument("configuration")
    parser.add_argument("backend", default="slurm", choices=("slurm", "dummy"),
                        nargs="?")

    # Prefer retaining immutable Arguments()
    # by not using the instance as a namespace
    return Arguments(**vars(parser.parse_args()))


def main():
    """CLI entry point
    """
    args = parse_args()

    if args.daemon:
        background_fork(True)

    setup_logging("{}".format(os.path.basename(args.configuration)),
                  verbose=args.verbose)

    logging.info("Model Ensemble Runner")

    config = EnsembleConfig(args.configuration)
    # TODO: get_batch_executor
    BatchExecutor(config,
                  args.backend,
                  dict(args.extra)).run()



import asyncio
import atexit
import collections
import concurrent.futures
import logging
import os
import random
import subprocess
import threading

from model_ensembler.cluster import Job, job_lock
from model_ensembler.utils import Arguments


START_STATES = ("SUBMITTED", "RUNNING")
FINISH_STATES = ("COMPLETED", "FAILED")

_dict_lock = threading.Lock()
_jobs = dict()


def threaded_job(run_dir, script):
    global _jobs

    with _dict_lock:
        _jobs[run_dir] = Job(_jobs[run_dir].name,
                             "RUNNING",
                             True,
                             False)

    logging.info("DUMMY RUN: {} - {}".format(run_dir, _jobs[run_dir]))
    subprocess.run("./{}".format(script), cwd=run_dir)

    # TODO: failed
    with _dict_lock:
        _jobs[run_dir] = Job(_jobs[run_dir].name,
                             "COMPLETED",
                             True,
                             True)


async def find_id(job_id):
    global _jobs

    job = None
    job_arr = [el for el in _jobs.values() if el.name == job_id]

    if len(job_arr) == 1:
        job = job_arr[0]
    elif len(job_arr) > 1:
        raise LookupError("{} jobs found for ID {}".format(len(job_arr),
                                                           job_id))
    return job


async def current_jobs(ctx, match):
    global _jobs

    job_arr = [el for el in _jobs.values()
               if el.name.startswith(match)
               and el.state in START_STATES]

    return job_arr


async def submit_job(ctx, script=None):
    # TODO: ugh, we could use contextvars for this
    global _jobs
    args = Arguments()

    max_submit_sleep = args.max_stagger
    sleep_for = random.randint(0, max_submit_sleep)
    logging.debug("Sleeping for {} seconds before submission".format(sleep_for))
    await asyncio.sleep(sleep_for)

    with _dict_lock:
        _jobs[ctx.dir] = Job(ctx.id, "SUBMITTED", False, False)

    threading.Thread(target=threaded_job, args=(ctx.dir, script)).start()
    return ctx.id

if __name__ == "__main__":
    pass



import importlib
import logging

from model_ensembler.tasks.utils import \
    check_task, processing_task, execute_command

# TODO: dynamic imports need to take place here
cluster = None


def init_hpc_backend(backend):
    global cluster

    logging.info("Importing {}".format(backend))
    try:
        cluster = importlib.import_module(backend)
    except (ImportError, ModuleNotFoundError) as e:
        logging.exception("Couldn't dynamically import cluster backend")
        raise e


"""HPC tasks

This module contains HPC related task methods
"""


async def find_id(job_id):
    return await cluster.find_id(job_id)


@check_task
async def jobs(ctx, limit, match):
    """Check: Assert whether number of jobs in SLURM is under limit

    Args:
        ctx (object): contextual configuration
        limit (int): number of jobs to check for
        match (str): prefix to match jobs by

    Returns:
        bool: true if number of jobs is less than limit, otherwise false
    """

    # TODO: match with regex
    async with cluster.job_lock:
        job_list = await cluster.current_jobs(ctx, match)
        res = len(job_list) < int(limit)

        logging.debug("Jobs in action {} with limit {}".format(
            len(job_list), limit))

    return res


@processing_task
async def submit(ctx, script=None):
    """Process: Submit a new job to SLURM

    Args:
        ctx (object): contextual configuration
        script (str, optional): slurm submission script for sbatch

    Returns:
        int: job identifier
    """

    # TODO: check this as an optional argument avoids run submission
    #  as intended
    if script:
        async with cluster.job_lock:
            await cluster.submit_job(ctx, script)

    return None


@check_task
async def quota(ctx, atleast, mnt=None):
    """Check: Make sure quota is sufficient

    Args:
        ctx (object): contextual configuration
        atleast (int): amount in kB
        mnt (str, optional): path for mount to check quota on if explicitly
        required

    Returns:
        bool: true if available space is less than atleast, false otherwise
    """

    # Command responds in 1k blocks
    path_arg = " -f " + mnt if mnt else ""
    quota_cmd = "quota -uw" + path_arg
    res = await execute_command(quota_cmd)
    quota_out = res.stdout.decode()

    try:
        fields = quota_out.splitlines()[-1].split()
        usage = int(fields[1])
        limit = int(fields[2])
        atleast = int(atleast)
    except (IndexError, TypeError):
        logging.exception("Could not reliably determine quota information")
        return False

    res = (limit - usage) >= atleast

    if not res:
        logging.warning("Quota remaining {} is less than {}".
                        format(limit - usage, atleast))
    return res



import asyncio
import functools
import inspect
import logging
import os
import types

from datetime import datetime

from model_ensembler.utils import Arguments

"""Task utilities

This module contains implementation and functionality relating generally to
tasks
"""


def flight_task(func, check=True):
    """Decorator for making func as a task, providing context preprocessing

    Args:
        func (callable): callable to wrap with context
        check (bool, optional): determine whether the func is to be treated
            as a check or another type of action (checks can be skipped)

    Returns:
        func: the wrapped function that can process the context provided
            appropriately
    """

    @functools.wraps(func)
    def new_func(ctx, *args, **kwargs):
        config = Arguments()

        for k, v in kwargs.items():
            try:
                if type(v) == str and str(v).startswith("run."):
                    nom = v.split(".")[1]
                    kwargs[k] = getattr(ctx, nom)
            finally:
                logging.debug("Calling with {} = {}".format(
                    k, kwargs[k])
                )

        # FIXME: context magic for execute_command calls below,
        #  not necessarily the best manner to achieve this as it would be
        #  better handled via the decorator
        if hasattr(ctx, 'dir') and 'cwd' in inspect.signature(func).parameters:
            kwargs['cwd'] = ctx.dir

        if config.no_checks and check:
            logging.info("Skipping checks for {}".format(func.__name__))
            return True
        return func(ctx, *args, **kwargs)

    new_func.check = check
    return new_func


check_task = functools.partial(flight_task, check=True)
processing_task = functools.partial(flight_task, check=False)


async def execute_command(cmd, cwd=None, log=False, shell=None):
    """Standard handling for calling external command

    Args:
        cmd (str): the relative path of the command being called to cwd
        cwd (str, optional): the current working directory to call the cmd
            from, passed to subprocess
        log (bool, optional): if true, output stdout/stderr to logfile in cwd
        shell (str, optional): which shell to ask subprocess to invoke when
            processing the command, will default to bash internally

    Returns:
        object: Namespace containing the returncode, stdout and stderr from
            the process that was invoked
    """

    logging.debug("Executing command {0}, cwd {1}".
                  format(cmd, cwd if cwd else "unset"))

    start_dt = datetime.now()

    args = Arguments()
    shell = args.shell if not shell else shell

    proc = await asyncio.create_subprocess_shell(
        cmd,
        executable=shell,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
        cwd=cwd)

    (stdout, stderr) = await proc.communicate()
    await proc.wait()

    if log and stdout:
        log_name = "execute_command.{}.log".\
            format(start_dt.strftime("%H%M%S.%f"))

        if cwd:
            log_name = os.path.join(cwd, log_name)

        with open(log_name, "w") as fh:
            fh.write(stdout.decode())

        logging.info("Command log written to {}".format(log_name))

    ret = types.SimpleNamespace(
        returncode=proc.returncode, stdout=stdout, stderr=stderr)

    if ret.returncode != 0:
        logging.warning("Command returned err: {}".format(ret.stderr))
        return ret
    else:
        logging.debug("Command successful")
    return ret



import logging
import os
import shutil

from .exceptions import FailureNotToleratedError
from .utils import check_task, processing_task, execute_command

"""System tasks

This module contains all methods for system related tasks
"""


@check_task
async def check(ctx, cmd, cwd=None, log=False, fail=False, shell=None):
    """Check: Call arbitrary command as a check

    Args:
        ctx (object): contextual configuration
        cmd (str): see ``utils.execute_command``
        cwd (str, optional): see ``utils.execute_command``
        log (bool, optional): see ``utils.execute_command``
        fail (bool, optional): if true, then the check returning nonzero will
            raise an error rather than return false, meaning run abandonment
            rather than recheck will take place
        shell (str, optional): see ``utils.execute_command``

    Returns:
        bool: true if return code is zero, false otherwise

    Raises:
        FailureNotToleratedError: error when fail is true and the check
            returns a nonzero code
    """

    logging.info("Running check: {}".format(cmd))

    res = await execute_command(cmd, cwd, log, shell)
    if res.returncode == 0:
        return True

    if fail:
        raise FailureNotToleratedError("Check failed and it is not tolerable")
    return False


@processing_task
async def execute(ctx, cmd, cwd=None, log=False, shell=None):
    """Process: Call arbitrary command as a processing task

    Args:
        ctx (object): contextual configuration
        cmd (str): see ``utils.execute_command``
        cwd (str, optional): see ``utils.execute_command``
        log (bool, optional): see ``utils.execute_command``
        shell (str, optional): see ``utils.execute_command``

    Returns:
        bool: true if return code is zero, false otherwise
    """
    logging.info("Running command: {}".format(cmd))

    res = await execute_command(cmd, cwd, log, shell)
    if res.returncode == 0:
        return True
    return False


# TODO: Context identification. WE MUST HAVE ID
@processing_task
async def move(ctx, dest, include=None, exclude=None, cwd=None):
    """Process: rsync current working directory contents

    Args:
        ctx (object): contextual configuration
        dest (str): path to copy ctx.id named directory to
        include (List[str], optional): rsync include specifiers
        exclude (List[str], optional): rsync exclude specifiers, defaults to
            "*" when calling rsync if include specifiers are given and no
            exclude specifiers are provided
        cwd (str, optional): see ``utils.execute_command``

    Returns:
        bool: true if return code is zero, false otherwise

    Raises:
        RuntimeError: did not provide necessary context attribute for using
            this processing task
    """

    if not hasattr(ctx, "id"):
        raise RuntimeError("No ID available for move")

    # TODO: Type checking
    include = [] if not include else include
    exclude = ["*"] if not exclude and include else []
    dest = os.path.join(dest, ctx.id)

    liststr = " ".join(["--include=\"{}\"".
                       format(i.strip("\"")) for i in include]) + " " + \
        " ".join(["--exclude=\"{}\"".
                 format(e.strip("\"")) for e in exclude])

    cmd = "rsync -aXE {} ./ {}/".format(liststr, dest)
    logging.info(cmd)

    res = await execute_command(cmd, cwd)
    if res.returncode == 0:
        return True
    return False


@processing_task
async def remove(ctx, directory=None):
    """Process: Remove directory using shutil.rmtree

    Args:
        ctx (object): contextual configuration
        directory (str, optional): specify the directory to remove, otherwise
            this will use the path specified by ctx.dir

    Returns:
        bool: true if return code is zero, false otherwise
    """
    if not directory:
        directory = ctx.dir

    logging.info("Attempting to remove data on {}".format(directory))

    try:
        shutil.rmtree(directory)
    except OSError as e:
        logging.exception("Could not remove {}: {}".
                          format(directory, e.strerror))
        return False
    return True



from setuptools import setup, find_packages

"""Setup module for model_ensembler

"""

from model_ensembler import __author__, __email__, __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="model-ensembler",
    version=__version__,
    author=__author__,
    author_email=__email__,
    description="Model Ensembler for managed batch workflows",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.github.com/jimcircadian/model-ensembler",
    project_urls={
        "Bug Tracker": "https://github.com/jimcircadian/model-ensembler/issues",
    },
    packages=find_packages(),
    keywords='slurm, hpc, tools, batch, model, ensemble, local',
    classifiers=[
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Development Status :: 3 - Alpha",
        "Topic :: System :: Distributed Computing",
    ],
    entry_points={
        'console_scripts': [
            "model_ensemble=model_ensembler.cli:main",
            "model_ensemble_check=model_ensembler.cli:check"
        ],
    },
    python_requires='>=3.7, <4',
    install_requires=[
        "jinja2",
        "jsonschema",
        "pyyaml",
    ],
    include_package_data=True,
)
